/* calculates value of PI
*   Input: None.
*   Output: Value of PI.
*   
*   Algorithm:
*   1. Each process calculates its interval of integration.
*   2. Each process estimates the integral of f(x) over its interval
*      using the trapezoidal rule.
*   3. Root gathers individual integral from each processor.
*   4. Root(processor 0) sums the calculations received from the 
*      individual processes & prints the result.

*   Author : Vidyut Tarkas
*            vtarkas@mailbox.syr.edu
*/

/***********************************************************

  Please  Run with 6 Nodes to get appropriate results

  **********************************************************/

#include <stdio.h>
#include <mpi.h>
#include <math.h>

int main(int argc,char *argv[])
{
  int my_rank;            /* My process rank          */
  int p;                  /* The number of processes  */
  double a = -0.5;         /* Left endpoint            */
  double b = 0.5;          /* Right endpoint           */
  int n = 1024;           /* Number of trapezoids     */
  double h;                /* Trapezoid base length    */
  double local_a;          /* Left endpoint for my process */
  double local_b;          /* Right endpoint for my process */
  int local_n;            /* Number of trapezoids for individual */
                          /* processes */    
  int source;             /* process sending integral  */
  double integral;         /* integral over my interval */
  double total;            /* Final PI value            */
  double arr_each_integral[];  /* array of individual integrals */

  double Trap(double local_a,double local_b,int local_n,double h);
  
  /* Initialise MPI */
  MPI_Init(&argc,&argv);
  
  /* Get my process rank */
  MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

  /* Find out how many processes are being used */
  MPI_Comm_size(MPI_COMM_WORLD,&p);

  h=(b-a)/n;
  local_n=n/p;
  
  /* Broadcast interval of integration for each processes */
  MPI_Bcast(&local_n, 1, MPI_INT, 0, MPI_COMM_WORLD);

  /* length of each process's interval of integration = local_n*h.
     so my interval starts at: */
  local_a=a+my_rank*local_n*h;
  local_b=local_a+local_n*h;
  integral=Trap(local_a,local_b,local_n,h);
  
  /* Gather the integrals calculated by each process */
  MPI_Gather(&integral,1,MPI_DOUBLE,arr_each_integral,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
  
  /* Add up the integrals calculated by each process */
  if(my_rank==0)
  {
     total=integral;
     for(source=1;source < p;source++)
         total+=arr_each_integral[source];
     printf("PI value is %lf\n",total);
  }

  /* Broadcast the final value of PI to each of the process */
  MPI_Bcast(&total, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD); 

  /* shut down MPI */
  MPI_Finalize();
}

double Trap(double local_a,double local_b,int local_n,double h)
{
   double integral;    /* store result in integral */
   double x;
   int i;
   
   double f(double x);  /* function we are integrating */
   
   integral=(f(local_a)+f(local_b))/2.0;
   x=local_a;
   /* compute the integral for subintervals */
   for(i=1;i<=local_n-1;i++)
   {
       x+=h;
       integral+=f(x);
   }
   integral*=h;
   return integral;
}

double f(double x)
{
  double return_val;
  return_val=4.0/(1.0+x*x);
  return return_val;
}

